open Format 

(* a set: a data-type with equality and input-output functions *)
module type Set =
sig
  type elem
    
  val (==) : elem -> elem -> bool     
      (* equality *)

  val normalize : elem -> elem      
      (* there may be more than one
	 representation of each element in the set,
	 this function may perform some simplification,
	 it is at least called by printing
      *)

  val print : elem -> unit          
      (* printing (using the Format library) *)
  val write : formatter -> elem -> unit    
      (* writing in file (using the Format library) *)
  val parse : char Stream.t -> elem
  val read : in_channel -> elem
      (* reading from file, compatible with write *)
  val write_bin : out_channel -> elem -> unit  
      (* writing as binary value *)
  val read_bin : in_channel -> elem         
      (* reading as binary value *)
end

module type Group =
sig
  include Set

  val zero : elem
    (* zero ! *)
  val (++) : elem -> elem -> elem         
      (* sum *)
  val (--) : elem -> elem -> elem         
      (* subtraction *)
  val opp : elem -> elem
      (* opposite (0 -- x) *)   
end

module type Ring =
sig
  include Group                         
    
  val one : elem                    
    (* one ! *)
  val t_of_int : int -> elem        
      (* the standard mapping from integer *)
  val ( ** ) : elem -> elem -> elem
      (* product *)
  val conjugate : elem -> elem
      (* conjugate value : usefull for complex numbre. It is 
         identity otherwise *)
end


module type Euclidian_Ring =
sig
  include Ring
      
  type stathm                        
    (* type of the stathm of an element of the ring :
       this is the type that is used to compare the size of the
       remaining in the division *)

  val stathm : elem -> stathm
      (* the stathm function *)
  val less : stathm -> stathm -> bool     
      (* comparison of stathm (strict) *)
      
  val (//) : elem -> elem -> elem
      (* euclidian division *)
  val (mod) : elem -> elem -> elem
      (* remaining of the euclidian division *)
  val div_mod : elem -> elem -> elem * elem
      (* div_mod x y = (x // y, x mod y) *)
end

module type Field =
sig
  include Ring
  val inv : elem -> elem             (* inverse in the filed *)
  val (//) : elem -> elem -> elem       (* division in the field *)
end

module type Normed_Field =
sig
  include Field
  type norm
    (* type of the norm of an element of the ring *)
    
  val norm : elem -> norm 
      (* the norm function (the square of the norm to be exact) *)
  val abs : elem -> norm
      (* absolute value. Should at least verify
	 norm x = abs (R.(**) (conjugate x) x), but could be the 
	 square root of the norm if it exists. *)
  val leq : norm -> norm -> bool
      (* comparison of norm *)
  val add_norm : norm -> norm -> norm
  val zero_norm : norm
end

module type Vectorial_Space =
sig
  include Group
  
  type scalar
  val ( @ ) : scalar -> elem -> elem 
  val conjugate : elem -> elem
      (* conjugate value *)
end
  
module type Banach =
sig
  include Group
  
  type scalar
  val ( @ ) : scalar -> elem -> elem 

  type norm
    (* type of the norm of an element of the ring *)

  val conjugate : elem -> elem
      (* conjugate value *)
  val norm : elem -> norm 
      (* the norm function (the square of the norm to be exact) *)
  val abs : scalar -> norm
      (* absolute value if possible. Should at least verify
      norm x = abs ((conjugate x) || x) *)  
  val (||) : elem -> elem -> scalar
      (* scalar product which is not always equal to the conjugate_product! *)
end

module type Mapping =
sig 
  type elem
  type argument
  type result
			    
  val apply : elem -> argument -> result (* mult by a vector  *)
end

(* integers in machine representation (type int : 32 or 64 bits) *)
module Ring_Z : Euclidian_Ring 
  with type elem = int and type stathm = int

(* multiprecision integers using the num library *)
module Ring_MZ : Euclidian_Ring 
  with type elem = Big_int.big_int and type stathm = Big_int.big_int

(* real number using the machine representation (type float) *)
module Field_R : Normed_Field 
  with type elem = float and type norm = float

(* multiprecision rational using the num library *)
module Field_Q : Normed_Field 
  with type elem = Num.num and type norm = Num.num

(* complex number using the machine representation (type float * float) *)
module Field_C : Normed_Field 
  with type elem = float * float and type norm = float

(* multiprecision complex rational using the num library *)
module Field_QC : Normed_Field 
  with type elem = Num.num * Num.num and type norm = Num.num

(* Missing Gauss integer ring ! *)

(* This functor qill give you an algorithm for pgcd for any
   Euclidian_Ring *)
module Pgcd :
  functor(R : Euclidian_Ring) ->
    sig
      val pgcd : R.elem -> R.elem -> R.elem
      val bezout : R.elem -> R.elem -> R.elem * R.elem * R.elem
    end

(* a signature that just hold one element of a type *)
module type One_element = 
sig 
  type elem
  val elt : elem 
end

(* quotient of a ring by an element *)
module Quotient :
  functor(R : Euclidian_Ring) ->
    functor(Elt : One_element with type elem = R.elem) ->
      Ring with type elem = R.elem
  
(* quotient of a ring by a prime element.
   you have to test the primality :-( *)
module Quotient_prime :
  functor(R : Euclidian_Ring) ->
    functor(Elt : One_element with type elem = R.elem) ->
      Field with type elem = R.elem
