(**************************************************
**             expression.ml                    ***
**          Formal calcul module                ***
**      POUSSARD PA    march 2000               ***
**************************************************)

open Input_util
open Algebra
open Format    

module type Expression  =
sig
  (* include Field *)
  type scalar

  type trans = (* transantes function, like sin() or cos() *)
      {
	name : string;
	calcul : scalar -> scalar;
	deriv : elem;
	name_var: string
      }
      
  and elem =
      Add of elem  * elem
    | Sub of elem  * elem
    | Div of elem  * elem
    | Mul of elem  * elem
    | Pow of elem  * int
    | Tra of trans * elem
    | Cst of scalar
    | Var of string
      
  type 'a environment = (string * 'a) list

(*** field struct ***)
  val zero : elem              (* zero ! *)
  val one :  elem              (* one ! *)
  val t_of_int : int -> elem   (* the standard mapping from integer *)
  val (++) : elem -> elem -> elem         
  val (--) : elem -> elem -> elem       
  val ( ** ) : elem -> elem -> elem      
  val ( @ ) : scalar -> elem -> elem      
  val (==) : elem -> elem -> bool      
  val opp : elem -> elem             
  val inv : elem -> elem
  val (//) : elem -> elem -> elem    
  val cst : scalar -> elem
  val var : string -> elem
  val trans : trans -> elem -> elem

  val print : elem -> unit          
  val write : formatter -> elem -> unit
  val parse : char Stream.t -> elem
  val read : in_channel -> elem    
  val write_bin : out_channel -> elem -> unit
  val read_bin : in_channel -> elem 
	
  val simplify : elem -> elem
  val normalize : elem -> elem
  val derive   : elem -> string -> elem
  val sder     : elem -> string -> int -> elem
  val eval     : elem -> scalar environment -> scalar	
  val conjugate : elem -> elem
end                                                                                       
module Expression =functor (R:Field) ->
struct
  type scalar = R.elem

  type trans = (* transantes function, like sin() or cos() *)
      {
	name : string;
	calcul : scalar -> scalar;
	deriv : elem;
	name_var: string
      }
      
  and elem =
      Add of elem  * elem
    | Sub of elem  * elem
    | Div of elem  * elem
    | Mul of elem  * elem
    | Pow of elem  * int
    | Tra of trans * elem
    | Cst of scalar
    | Var of string
	
  type 'a environment = (string * 'a) list
			  
  type  expS = (* simplify type*)
      Som of (scalar * expS) list
    | Pro of (int * expS) list
    | Tra' of trans * expS
    | Var' of string
    | Cst' of scalar
	
    let rec eval_power exp puis = 
       (* power function in field (time log_2(n))*)
       match puis with
	 0 -> R.one
       | 1 -> exp
       | _ when puis < 0 -> R.inv(eval_power exp (-puis))
       | _ when puis mod 2 = 0 ->
	   let r = eval_power exp (puis / 2) in
	   R.( ** ) r r
       | _ -> 
	   let r = eval_power exp ((puis - 1) / 2) in
 	   R.( ** ) exp (R.( ** ) r r) 
	  
      (******** Simplification procedure *******)
	  
  let rec totypeS e =
    match e with
      Cst n -> Cst' n
    | Var var -> Var' var
    | Tra(tra,exp')->Tra'(tra,totypeS exp')
    | Pow(exp,puis)->Pro([puis,totypeS exp])
	(* (R.one,cte R.one est la pour que l'expression est deux argument *)
    | Mul(e1,e2) -> Pro([1,totypeS e1 ; 1,totypeS e2])
    | Div(e1,e2) -> Pro([1,totypeS e1 ; -1,totypeS e2])
    | Add(e1,e2) -> Som([R.one,totypeS e1 ; R.one,totypeS e2])
    | Sub(e1,e2) -> Som([R.one,totypeS e1 ; R.opp(R.one),totypeS e2])
	
  let rec totypeE e   = 
    match e with 
      Cst' n -> Cst n
    | Var' var-> Var var 
    | Tra'(tra, exp')->Tra(tra,totypeE exp')
    | Pro([]) -> Cst R.one
    | Pro(x::l)-> 
	let fn y = match y with
	    (1,exp)-> totypeE exp
	  | (n,exp) when exp=Cst' R.one -> Cst R.one
	  | (n,exp)-> Pow(totypeE exp,n) in
	let f exp arg = Mul(exp, fn arg) in
	List.fold_left f (fn x) l
    | Som([]) -> Cst R.zero   
    | Som(x::l)-> 
	let fn y = match y with
	    (a,exp) when a = R.one -> totypeE exp
	  | (n,Cst' a) -> Cst( R.( ** ) n a )
	  | (n,exp)-> Mul(Cst n, totypeE exp) in
	let f exp arg = Add(exp, fn arg) in
	List.fold_left f (fn x) l

  (* ordre sur les termes, utilise pour orienter les sommes et les produits *)
  let prior = function
      Cst' _ -> 0
    | Var' _ -> 1
    | Som  _ -> 2
    | Pro  _ -> 3
    | Tra' _ -> 4
	
	
  let rec cmp_list l l' fn = match l, l' with
      [], [] -> 0
    | [], _::_ -> -1
    | _::_, [] -> 1
    | (a,x)::l, (a',x')::l' ->
	match fn x x' with
	  0 -> 
	    begin
	      match cmp_list l l' fn with
		0 -> compare a a' 
	      | c -> c
	    end	  	 
	| c -> c	  
	    
  let rec cmp' x y = match x,y with
      Cst' n, Cst' n' -> compare n n' 
    | Var' name, Var' name' -> compare name name'
    | Som (l), Som(l') -> cmp_list l l' cmp'
    | Pro (l), Pro(l') -> cmp_list l l' cmp'
    | Tra'({name = s},l),Tra'({name = s'},l') -> 
	begin
	  match compare s s' with
	    0 -> cmp' l l'
	  | c -> c
	end
    | l1, l2 -> compare (prior l1) (prior l2)
	
	
  let cmpBool x y = cmp'(snd x)(snd y) <= 0
		      
  let rec regroupSom  list  = match list with
      (a,x)::l when a = R.zero  -> regroupSom  l 
    | (c,Cst' n)::l when n= R.zero   -> regroupSom  l
    | (c,x)::(c',x')::l when cmp' x x' = 0 -> 
	regroupSom  ((R.(++) c c', x)::l) 
    | (c,x)::l -> (c,x)::regroupSom  l 
    | [] -> []
	
  let rec regroupPro list  = match list with
      (a,_)::l when a = 0  -> regroupPro l 
    | (_,Cst' n)::l when n= R.one   -> regroupPro l
    | (_,Cst' n)::l when n= R.zero   -> [1,Cst' R.zero]
	(* cas on ne peut pas tout passer dans le premier membre comme pour l'add *)
    | (c,Cst' n)::(c',Cst' n')::l -> 
	regroupPro ((1,Cst' (R.( ** ) 
			       (eval_power n c) 
			       (eval_power n' c')
			    ))::l)
    | (c,x)::(c',x')::l when (cmp' x x') = 0 -> 
	regroupPro (( c + c', x)::l) 
    | (c, Cst' n)::l -> (1,Cst' (eval_power n c))::regroupPro l
    | (c,x)::l -> (c,x)::(regroupPro l) 
    | [] -> []
	
  let rec simplS expS = (*simplification fonction with type expS*)
	let rec simplSND l =
	  (fst l,simplS (snd l)) 
	in
	match expS with
	  Cst' n -> Cst' n
	| Var' name  -> Var' name
	| Tra'(tra,exp) -> Tra'(tra,simplS exp)
	| Pro(l) -> 
	    let l = List.map simplSND l in
	    let rec fn = function
		(c,Pro l) -> 
		  List.map (fun (c', x) -> (c * c', x)) l
	      | (c, Som []) ->
		  [c, Cst' R.zero]
	      | (c, Som [c', e]) ->
		  (c, Cst' c')::fn (c,e)
	      | (c, x) ->  
		  [c, x]
	    in
	    let l = List.flatten (List.map fn l) in
	    let l = Sort.list (cmpBool) l in
	    Pro(regroupPro l)	  
	| Som(l) -> 
	    let l = List.map simplSND l in
	    let rec fn = function
		(c,Som l) -> 
		  List.map (fun (c', x) -> (R.( ** ) c c', x)) l
	      | (c, Pro []) ->
		  [c, Cst' R.one]
	      | (c, Pro([a,Cst' n])) ->
		  [R.( ** ) c (eval_power n a), Cst' R.one]
	      | (c, Pro([1,e])) ->
		  fn (c, e)
	      | (c, Pro((a,Cst' n)::l1)) ->
		  fn (R.( ** ) c (eval_power n a), Pro l1)
	      | (c, x) ->   
		  [c, x]
	    in
	    let l = List.flatten (List.map fn l) in
	    
	    let fncst = function 
		(c,Cst' n) -> (R.( ** ) c n , Cst' R.one)
	      | (c,a) -> (c,a) in
	    let l = List.map fncst l in
	    
	    let l = Sort.list (cmpBool) l in
	    Som(regroupSom l)	  
	      
  let simplify exp = 
    totypeE (simplS (totypeS exp))
      
  let normalize = simplify

      (************ field's definition ***************)	
      
  let zero = Cst R.zero 
  let one  = Cst R.one                              
  let t_of_int n = Cst (R.t_of_int n)
  let (++)  exp1 exp2 = Add(exp1,exp2)
  let (--)  exp1 exp2 = Sub(exp1,exp2)
  let (//) exp1 exp2 = Div(exp1,exp2)
  let ( ** ) exp1 exp2 = Mul(exp1,exp2)
  let (@) x e = Mul(Cst x, e)
  let opp exp= Sub(exp,Cst R.zero)
  let inv exp =Div( Cst R.one, exp) 
  let cst c = Cst c
  let var s = Var s
  let trans f e = Tra(f, e)

	
  let (==) exp1 exp2 = 
    let rec fn x y = match x,y with
	Cst n, Cst n' -> R.(==) n n' (* compare with lexicographic ordre *)
      | Var name, Var name' -> name = name'
      | Add(e1,e2), Add(e1',e2') -> fn e1 e1' && fn e2 e2'
      | Sub(e1,e2), Sub(e1',e2') -> fn e1 e1' && fn e2 e2'
      | Mul(e1,e2), Mul(e1',e2') -> fn e1 e1' && fn e2 e2'
      | Div(e1,e2), Div(e1',e2') -> fn e1 e1' && fn e2 e2'
      | Pow (e,n),  Pow (e',n') ->  fn e e' && n = n'
      | Tra ({name = s},e), Tra ({name = s'},e') -> s = s' && fn e e'
      | e1, e2 -> false
    in fn (simplify exp1) (simplify exp2)
      
  let eval expr v = (* eval expression in environnent *)
    let rec fn = function
	Var name -> List.assoc name v 
      | Cst n -> n
      | Add (e,e') -> R.(++) (fn e) (fn e')
      | Sub (e,e') -> R.(--) (fn e) (fn e')
      | Mul (e,e') -> R.( ** ) (fn e) (fn e')
      | Div (e,e') -> R.( // ) (fn e) (fn e')
      | Pow (e,deg) -> eval_power (fn e) deg 
      | Tra ({calcul = f}, e) -> f (fn e)
    in fn expr
	 
      (* substitution "name" var in "exp1" with "exp2"
	 fr= substitution d'une variable "name" par une expression "exp2" dans la formule "exp1" *) 
  let subst exp1 name exp2 =
    let rec fn = function
	Var name' -> if name = name' then exp2 else Var name'
      | Cst _ as e -> e
      | Add (e,e') -> Add (fn e, fn e')
      | Sub (e,e') -> Sub (fn e, fn e')
      | Mul (e,e') -> Mul (fn e, fn e')
      | Div (e,e') -> Div (fn e, fn e')
      | Pow (e,n) ->  Pow (fn e, n)
      | Tra (tran,e) -> Tra (tran, fn e)
    in fn exp1
	 
  let conjugate exp1 =
    let rec fn = function
	Var _ as e -> e
      | Cst (v) -> Cst (R.conjugate v)
      | Add (e,e') -> Add (fn e, fn e')
      | Sub (e,e') -> Sub (fn e, fn e')
      | Mul (e,e') -> Mul (fn e, fn e')
      | Div (e,e') -> Div (fn e, fn e')
      | Pow (e,n) ->  Pow (fn e, n)
      | Tra (tran,e) -> Tra (tran, fn e)
    in fn exp1

      (********** output input functions ***********)
	 
      (* la fonction d'affichage, utilisant les priorits usuelles pour les oprateurs binaires *)
  let write formatter  expr = 
    
    let print_string s = 
      pp_print_string formatter s
    in
    let print_string' s = 
      print_space (); 
      pp_print_string formatter s; 
      print_space ()
    in

     (* dans cette fonctions auxilliaire, le paramtre "l" reprsente le niveau de priorit courant *)
    let rec fn l = function
	Var name -> print_string name
      | Cst n -> R.print n
      | Add _ | Sub _ as e -> pp l 1 (fun () -> fns 1 e) ()
      | Mul _ | Div _ as e -> pp l 2 (fun () -> fnp 2 e) ()
      | Pow (e,n) ->  
	      pp l 3 (fun () -> fn 3 e) (); print_string "^"; print_int n 
      | Tra ({name = s},e) -> print_string s; print_string "(";
          fn 0 e; print_string ")"
	    
	(* affichage associatif des sommes *)
    and fns l = function
      Add(e,e') -> fns l e; print_string' "+"; fns l e'
    | Sub(e,e') -> fns l e; print_string' "-"; fn (l+1) e';
    | e -> fn l e
	
	(* affichage associatif des produits *)
    and fnp l = function
      Mul(e,e') -> fns l e; print_string' "*"; fns l e'
    | Div(e,e') -> fns l e; print_string' "/"; fn (l+1) e'
    | e -> fn l e
	
	(* fonctions testant les priorits pour afficher ou non les paranthses *)
    and pp l l' f e =
      if l > l' then print_string "("; 
      pp_open_box formatter 2;
      f e;
      pp_close_box formatter ();
      if l > l' then print_string ")"

    in 
    pp_open_box formatter 2;
    fn 0 expr;
    pp_close_box formatter ()
	
	 
  let write_bin ch x = output_value ch x
  let read_bin ch = input_value ch     
  let print = write std_formatter

  let trans_table = ref [] 
		      
  let rec parse str = 
    parse_spaces str; 
    match str with parser
	[< e1 = parse_mult;e = parse_more_adds e1 >] -> e
      | [< ''-'; e1 = parse_mult;e = parse_more_adds (Sub(zero,e1)) >] -> e
	  
  and parse_more_adds e1 str =        
    parse_spaces str; 
    match str with parser
	[< ''+'; e2 = parse_mult;e = parse_more_adds (Add(e1,e2))>] -> e
      | [< ''-'; e2 = parse_mult;e = parse_more_adds (Sub(e1,e2))>] -> e
      | [< >] -> e1
	  
  and parse_mult str = 
    parse_spaces str; 
    match str with parser
	[< e1 = parse_power;e = parse_more_mults e1 >] -> e
	    
  and parse_more_mults e1 str = 
    parse_spaces str; 
    match str with parser
	[< ''*'; e2 = parse_power;e = parse_more_mults (Mul(e1,e2))>] -> e
      | [< ''/'; e2 = parse_power;e = parse_more_mults (Div(e1,e2))>] -> e
      | [< >] -> e1
	  
  and parse_power str =
    parse_spaces str; 
    match str with parser
	[< e1 = parse_simple; e = parse_more_power e1 >] -> e
	    
  and parse_more_power e1 str =
    parse_spaces str; 
    match str with parser
	[< ''^'; e2 = parse_int; 
	   e = parse_more_power (Pow(e1,e2)) >] -> e
      | [< >] -> e1
	  
  and parse_trans name str = 
    parse_spaces str; 
    match str with parser
      | [< ''('; e2 = parse; _ = parse_spaces;  '')' >] -> 
	  begin
	    try 
	      Tra(List.assoc name !trans_table, e2)
	    with Not_found -> 
	      raise (Stream.Error ("unbound function: "^name))
	  end
      | [< >] -> Var(name)
	  
  and parse_simple str = 
    parse_spaces str; 
	match str with parser
	  | [< ''('; e = parse; _ = parse_spaces; '')' >] -> e 
	  | [< c = R.parse >] -> Cst(c)
	  | [< name = parse_ident; e = parse_trans name >] -> e


  let read ch = parse (Stream.of_channel ch)

      (* derivation par rapport a la varible name *) 
	    
  let derive expr name =
    let rec fn = function
	Var name' -> if name = name' then Cst R.one else Cst R.zero
      | Cst _ -> Cst R.zero
      | Add (e,e') -> Add (fn e, fn e')
      | Sub (e,e') -> Sub (fn e, fn e')
      | Mul (e,e') -> Add (gn e e', gn e' e)
      | Div (e,e') -> Sub (gn (Div(Cst R.one,e')) e, gn (Div(e, Pow (e',2))) e')
      | Pow (e,n) -> Mul (Cst (R.t_of_int(n)), Mul (fn e, Pow (e,(n-1)))) 
      | Tra (tran,e) -> Mul ((subst tran.deriv tran.name_var e), fn e)
    and gn acc = function
    | Mul (e,e') -> 
	  Add (gn (Mul(acc, e')) e, gn (Mul(acc, e)) e')
      | Div (e,e') -> 
	  Sub(gn (Div(acc, e')) e, gn (Mul(acc, Div(e, Pow (e',2)))) e')
      | e ->
	  Mul(acc, fn e)

    in simplify(fn expr)
	 
  let rec sder fn var  = function
      0 -> fn
    | n -> sder (derive fn var) var (n-1)

end

