let rec parse_spaces str = match str with parser 
  [< '' '|'\n'|'\r'|'\t'; _ = parse_spaces >] -> ()
| [< >] -> ()

let rec parse_delim op cl f str =
  parse_spaces str;
  match str with parser
    [< 'c when c = op; 
       r = parse_delim op cl f; 
       _ = parse_spaces; 'c when c = cl >] -> r
  | [< r = f >] -> r

let buf = String.create 1024

let rec parse_digits' pos str = match str with parser 
    [< ''0'..'9' as c; npos = parse_digits' (pos + 1) >] ->
      buf.[pos] <- c; npos
  | [< >] -> pos
let parse_digits pos str = match str with parser 
    [< ''0'..'9' as c; npos = parse_digits' (pos + 1) >] ->
      buf.[pos] <- c; npos

let parse_int' str = 
  let npos = match str with parser 
      [< ''+'; _ = parse_spaces; npos = parse_digits 0 >] -> npos
    | [< ''-'; _ = parse_spaces; npos = parse_digits 1 >] -> 
	buf.[0] <- '-'; npos
    | [< npos = parse_digits 0 >] -> npos
  in
  int_of_string (String.sub buf 0 npos)

let parse_int = parse_delim '(' ')' parse_int' 

let parse_big_int' str = 
  let npos = match str with parser 
      [< ''+'; _ = parse_spaces; npos = parse_digits 0 >] -> npos
    | [< ''-'; _ = parse_spaces; npos = parse_digits 1 >] -> 
	buf.[0] <- '-'; npos
    | [< npos = parse_digits 0 >] -> npos
  in
  Big_int.big_int_of_string (String.sub buf 0 npos)

let parse_big_int = parse_delim '(' ')' parse_big_int' 

let parse_float_exp' pos str = 
  buf.[pos] <- 'e';
  let pos = pos+1 in
  match str with parser 
    [< npos = parse_digits pos >] -> npos
  | [< ''+'; npos = parse_digits pos >] -> npos
  | [< ''-'; npos = parse_digits (pos+1) >] -> buf.[pos] <- '-'; npos

let parse_float_exp pos str = 
  if pos = 0 then match str with parser
    [< ''e'|'E'; npos = parse_float_exp' pos >] -> npos
  else match str with parser
    [< ''e'|'E'; npos = parse_float_exp' pos >] -> npos
  | [< >] -> pos

let parse_float_dec pos str = match str with parser 
    [< ''.'; pos' = parse_digits' (pos+1); 
       npos = parse_float_exp pos' >] -> buf.[pos] <- '.'; npos
  | [< npos = parse_float_exp pos >] -> npos

let parse_float' str = 
  let npos = match str with parser 
      [< pos = parse_digits 0; 
	 npos = parse_float_dec pos >] -> npos   
    | [< ''+'; _ = parse_spaces; pos = parse_digits' 0; 
	 npos = parse_float_dec pos >] -> npos
    | [< ''-'; _ = parse_spaces; pos = parse_digits' 1; 
	 npos = parse_float_dec pos >] -> buf.[0] <- '-'; pos
    | [< npos = parse_float_dec 0 >] -> npos
  in      
  float_of_string (String.sub buf 0 npos) 

let parse_float = parse_delim '(' ')' parse_float' 

let parse_denom str =
  parse_spaces str;
  match str with parser 
      [< ''/'; _ = parse_spaces; n = parse_big_int >] -> Num.Big_int n
    | [< >] -> Num.Int 1

let parse_num' str =
  match str with parser 
      [< n = parse_big_int; p = parse_denom >] -> Num.div_num (Num.Big_int n) p

let parse_num = parse_delim '(' ')' parse_num' 

let parse_tuple' f g str =
  match str with parser 
      [< fst = f; _ = parse_spaces; '','; 
	 _ = parse_spaces; snd = g >] -> fst, snd

let parse_tuple f g str = parse_delim '(' ')' (parse_tuple' f g) str 

let parse_complex zero opp f str =
  let fn str = match str with parser 
      [< ''+'; _ = parse_spaces; re = f >] -> re
    | [< ''-'; _ = parse_spaces; re = f >] -> opp re
    | [< >] -> zero
  in
  let gn str = match str with parser 
      [< ''+'; _ = parse_spaces; ''i'; _ = parse_spaces; im = f >] -> im
    | [< ''-'; _ = parse_spaces; ''i'; _ = parse_spaces; im = f >] -> opp im
    | [< >] -> zero
  in
  match str with parser 
      [< ''i'; _ = parse_spaces; im = f; 
	 _ = parse_spaces; re = fn >] -> re, im
    | [< re = f; _ = parse_spaces; im = gn >] -> re, im

let rec parse_alpha' pos str = match str with parser 
    [< ''0'..'9'|'_'|'a'..'z'|'A'..'Z' as c; 
       npos = parse_alpha' (pos + 1) >] ->
      buf.[pos] <- c; npos
  | [< >] -> pos

let parse_alpha pos str = match str with parser 
    [< ''a'..'z'|'A'..'Z' as c; npos = parse_alpha' (pos + 1) >] ->
      buf.[pos] <- c; npos

let parse_ident' str = 
  let npos = match str with parser 
      [< npos = parse_alpha 0 >] -> npos
  in
  String.sub buf 0 npos

let parse_ident = parse_delim '(' ')' parse_ident' 
