open Format
open Input_util
open Algebra
open Vector


(* Signature of light matrix. 
   Provide the minimum operation to implement iterative solver. *)
(* This is a subtype of Set and Mapping *)

module type Light_Matrix =
sig 
  include Set
    
  type scalar
  type argument
  type result

  (* direct acces to the number of columns and lines *)
  val columns : int (* Domain.dim *)
  val lines : int   (* Image.dim  *)

  (* product A * v *) 
  val apply : elem -> argument -> result

  (* product v * A = tranpose(A) * v *)

  val tapply : result -> elem -> argument

  (* universal creation : to use write a code like

     let init f =
       ...
       f i j x (* called if x is at position (i,j) for all non zero element *)
               (* i is the column number and j is the line number *)
       ...
     in
     let m = create init
  *) 
  val create : ((int -> int -> scalar -> unit) -> unit) -> elem 

  val copy : elem -> elem (* copy in an other matrix *)
end

(* Signature of matrix. 
   Provide everething but product (this is only for square matrix !). 
   Useful: Matrix are a subtype of Vector ! *)

module type Matrix =
  sig
    include Light_Matrix
 
    module D : Dim

    val dim : int
    val get : elem -> int -> int -> scalar
    val put : elem -> int -> int -> scalar -> unit
    val vget : elem -> int -> scalar
    val vput : elem -> int -> scalar -> unit

    val sub : elem -> elem -> unit (* substraction: m1 = m1 - m2 *)
    val add : elem -> elem -> unit (* sum: m1 = m1 + m2 *)
    val mul_scal : scalar -> elem -> unit (* m1 = a * m1 *)
	
    val (++) : elem -> elem -> elem (* sum *)
    val (--) : elem -> elem -> elem (* substraction *)
    val (@)  : scalar -> elem -> elem (* multiplication by a scalar *)

    (* the same, but create a matrix as if it were a vector 
       of size dim = lines * columns ! *)
    val vcreate : ((int -> scalar -> unit) -> unit) -> elem 

  end

(* square matrix: this time we have a product
   Square_Matrix are a subtype of Ring ! *)

module type Square_Matrix =
  sig
    include Matrix (* with type argument = result, it is a pity we can not
                      write this in Ocaml ! *) 

    val ( ** ) : elem -> elem -> elem
    val one : elem
    val t_of_int : int -> elem
  end 

(* Product of two matrix as a functor !*)

module Product : 
  functor (R : Ring) ->
  functor (ML : Matrix with type scalar = R.elem)  ->
  functor (MR : Matrix with type scalar = R.elem
                       and  type argument = ML.result) ->
  functor (M  : Matrix with type scalar = R.elem
                       and  type argument = MR.argument
                       and  type result = ML.result) ->
  sig
    val ( ** ) : ML.elem -> MR.elem -> M.elem
  end
