open Format
open Algebra
open Expression
open Vector
open Matrix
open Full_matrix

module Dim = struct let dim = 3 end

module V = Vector.Make_Normed(Field_R)(Dim)

module M = Make_Square(Field_R)(V)

module E = Expression(Field_R)

let x = E.Var "x"
let y = E.Var "y"
let z = E.Var "z"
let x2 = E.Pow(x, 2)
let y2 = E.Pow(y, 2)
let z2 = E.Pow(z, 2)
let x3 = E.Pow(x, 3)
let y3 = E.Pow(y, 3)
let z3 = E.Pow(z, 3)
let sphere = E.(++) (E.( ** ) x2 (E.Cst 2.0)) 
	       (E.(++) (E.( ** ) y2 (E.Cst 2.0))  
		  (E.(--) (E.( ** ) z2 (E.Cst 2.0)) (E.Cst 1.10))) 
let cubic = E.(++) x3 (E.(++) y3 (E.(++) z3 sphere)) 
let tore = E.(--) (E.Pow(E.(++) x2 (E.(++) y2 (E.(++) z2 (E.Cst 0.75))),2))
              (E.( ** ) (E.Cst 4.0) (E.(++) x2 y2))
let planes = E.( ** ) (E.(--) (E.(++) x y) (E.Cst 1.0)) (E.(--) (E.(++) x y) (E.Cst 2.0))
let e = cubic
let ex = E.derive e "x" 
let ey = E.derive e "y" 
let ez = E.derive e "z" 

(* let e = sphere  *)
let max_depth = 29
let min_depth_search = 3
let min_depth = 5
let size = 8.0
let hsize = size /. 2.0
let origin = V.from_array [| -.hsize; -.hsize; -.hsize |]

let eval e v = E.eval e ["x", V.vget v 0; "y", V.vget v 1; "z", V.vget v 2]  

let epsilon = 10e-10
let epsilon2 = 10e-14

let middle a b = V.map2 (fun x y -> (x +. y) /. 2.0) a b 

let dighotomie f a b =
  let fa = eval f a and fb = eval f b in
  if abs_float(fa) < epsilon then a else 
  if abs_float(fb) < epsilon then b else
  let fa, a, fb, b =     
    if fa < 0.0 && fb > 0.0 then fa, a, fb, b else
    if fa > 0.0 && fb < 0.0 then fb, b, fa, a else
    raise Not_found
  in
  let rec fn fa a fb b =
    let c = middle a b in
    let fc = eval f c in 
    if abs_float(fc) < epsilon then c else
    if fc < 0.0 then fn fc c fb b else fn fa a fc c 
  in
  fn fa a fb b

type index = int array

let step p = 1 lsl (max_depth - p)

type cube_vertex =
{
  index : index;
  cpos : V.elem;
  cfval : float;
  cube_edges : cube_edge option array;
}

and cube_edge =
{
  cvertex1 : cube_vertex;
  cvertex2 : cube_vertex;
  rvertex : vertex option;
  depth : int;
}

and vertex = {
  pos : V.elem;
  normal : float * float * float;
}

let build_vertex pos = {
  pos = pos;
  normal = (eval ex pos,eval ey pos,eval ez pos)
}

let directioni = [|
  [| 0 ; 0 ; 0 |];
  [| 1 ; 0 ; 0 |];
  [| 0 ; 1 ; 0 |];
  [| 0 ; 0 ; 1 |];
  [| 1 ; 1 ; 0 |];
  [| 0 ; 1 ; 1 |];
  [| 1 ; 0 ; 1 |];
  [| 1 ; 1 ; 1 |];
|]

let directions = [|
  V.from_array [| 0.0 ; 0.0 ; 0.0 |];
  V.from_array [| 1.0 ; 0.0 ; 0.0 |];
  V.from_array [| 0.0 ; 1.0 ; 0.0 |];
  V.from_array [| 0.0 ; 0.0 ; 1.0 |];
  V.from_array [| 1.0 ; 1.0 ; 0.0 |];
  V.from_array [| 0.0 ; 1.0 ; 1.0 |];
  V.from_array [| 1.0 ; 0.0 ; 1.0 |];
  V.from_array [| 1.0 ; 1.0 ; 1.0 |];
|]

let edges = [| 0,1,0,1; 0,2,2,3; 0,3,4,5; 1,4,2,3; 1,6,4,5; 2,4,0,1;
               2,5,4,5; 3,5,2,3; 3,6,0,1; 4,7,4,5; 5,7,0,1; 6,7,2,3 |]                                       
let all_vertex = Hashtbl.create 101

let all_cubes = ref []


let position index =
  V.vcreate (fun fn ->
    for i = 0 to 2 do
      fn i (V.vget origin i +. size *. (float) index.(i) /. 
           (float) (1 lsl max_depth))
    done)

let rec do_edge fn v1 v2 dir =
  let Some e = v1.cube_edges.(dir) in
  let v = 
    if e.cvertex1 == v1 then e.cvertex2 else
    if e.cvertex2 == v1 then e.cvertex1 else
    failwith "bug in do_edge"
  in
  fn e; 
  if not (v == v2) then do_edge fn v v2 dir

let split_condition depth cube =
  let non_empty = 
    let fn e = if e.rvertex <> None then raise Exit in
    try
      for i = 0 to 11 do
	let vi1, vi2, i1, i2 = edges.(i) in
	let v1 = cube.(vi1) and v2 = cube.(vi2) in
	do_edge fn v1 v2 i1
      done;
      raise Not_found
    with
    | Not_found -> false
    | Exit -> true
  in  
  if depth >= max_depth then non_empty, false else
  if depth < min_depth_search then non_empty, true else
  if depth < min_depth && non_empty then non_empty, true else
  non_empty, false

let rec make_cube e index depth =
  let fn i = 
    let index = Array.init 3 (fun j -> index.(j) + directioni.(i).(j) * step depth) in
    try 
      Hashtbl.find all_vertex index 
    with Not_found ->
      let pos = position index in
      let rec adjust pos =
        let x =  eval e pos in
	if x = 0.0 then 
	  let npos = V.map (fun x -> x +. size *. Random.float epsilon2) pos in
	  adjust npos
	else
	  pos, x 
      in
      let pos, x = adjust pos in
      let vertex = { index = index; cpos = pos; cfval = x; 
	cube_edges = Array.make 8 None } 
      in
      Hashtbl.add all_vertex index vertex;
      vertex
  in 
  let cube = Array.init 8 fn in
  for i = 0 to 11 do
    let vi1, vi2, i1, i2 = edges.(i) in
    let v1 = cube.(vi1) and v2 = cube.(vi2) in
    match v1.cube_edges.(i1), v2.cube_edges.(i2) with
    | Some e1, Some e2 when e1.depth > depth && e2.depth > depth ->
	()
    | Some e1, Some e2 when e1.depth < depth && e2.depth >= depth ->
			 failwith "bug 1 in make_edge"
    | Some e1, Some e2 when e1.depth >= depth && e2.depth < depth ->
			 failwith "bug 2 in make_edge"
    | _ ->
	let edge = {cvertex1 = v1; cvertex2 = v2; depth = depth; rvertex =
		      if v1.cfval *. v2.cfval < 0.0 then
			let x = dighotomie e v1.cpos v2.cpos in
			Some (build_vertex x)
		      else
			None }
	in v1.cube_edges.(i1) <- Some edge; v2.cube_edges.(i2) <- Some edge; 
  done;
  let non_empty, split = split_condition depth cube in
  if split then
    for i = 0 to 7 do
      let index = Array.init 3 (fun j -> index.(j) + directioni.(i).(j) * step (depth + 1)) in
      make_cube e index (depth + 1)
    done
  else if non_empty then all_cubes := cube :: !all_cubes

let _ = make_cube e [|0;0;0|] 0

let draw_surface e = 
  let list = GlList.create `compile in
  
  let vertex a =
    GlDraw.normal3 (a.normal);
    GlDraw.vertex3 (V.vget a.pos 0,V.vget a.pos 1,V.vget a.pos 2)
  in
  let triangle a b c =
    GlDraw.begins `triangle_strip;
    vertex a; vertex b; vertex c;
    GlDraw.ends ()
  in
  let rec triangularize l =
    match l with
      x::y::z::l -> triangle x y z; triangularize (x::z::l); triangularize (z::l@[x])
    | _ -> ()
  in

  let make_triangles cube =
    let composantes = Array.init 8 (fun i -> i) in
    let rec search i = 
      let i' = composantes.(i) in
      if i' <> i then search i' else i
    in
    let merge i j =
      let i' = search i and j' = search j in
      if i' <> j' then composantes.(i') <- j' 
    in
    for i = 0 to 11 do
      let vi1, vi2, i1, i2 = edges.(i) in
      let v1 = cube.(vi1) and v2 = cube.(vi2) in
      let Some e = v1.cube_edges.(i1) in
      if e.rvertex = None then merge vi1 vi2
    done;
    let composantes_lists = ref [] in
    for i = 0 to 7 do
      if search i = i then begin
	let cur = ref [] in
	for j = 0 to 7 do
	  if search j = i then cur := j::!cur
	done;
	composantes_lists := !cur::!composantes_lists
      end;
    done;
    let extract = function
	Some e -> e
      | None -> raise Not_found
    in
    let add x l = if not (List.memq x !l) then l := x::!l in
    let search l cond =
      let rec fn = function
          y::l -> 
	    if cond y then y
	    else fn l
	| [] -> raise Not_found
      in fn l
    in
    let same_face i j =
      let [|i0;i1;i2|] = cube.(i).index in
      let [|j0;j1;j2|] = cube.(j).index in
      (i0 = j0 or i1 = j1 or i2 = j2)
    in
    let last_dir = ref 0 in
    let common_edge i j =
      let [|i0;i1;i2|] = cube.(i).index in
      let [|j0;j1;j2|] = cube.(j).index in
      let dir = match i0 - j0, i1 - j1, i2 - j2 with
	  x, 0, 0 when x != 0 -> if x < 0 then 0 else 1
	| 0, x, 0 when x != 0 -> if x < 0 then 2 else 3
	| 0, 0, x when x != 0 -> if x < 0 then 4 else 5
	| _ -> -1 
      in
      last_dir := dir; dir >= 0
    in 
    let build_polygone c c' = 
      try
	let last_result = ref None in
	let x = search c 
		  (fun x ->
		     try
		       let y = search c' (common_edge x) in
		       last_result := Some y; true
		     with Not_found -> false) in
	let Some y = !last_result in
	let Some { rvertex = Some pt } = cube.(x).cube_edges.(!last_dir) in
	let rec fn adone acc x y =
	  print_int x; print_int y; print_string " ";
	  try 
	    let last_result = ref None in
	    let x1 = search c 
		       (fun x1 ->
			  same_face x x1 && same_face y x1 &&
			  try
			    let y1 = search c' 
				       (fun y1 ->
					  not (List.mem (x1, y1) adone) && common_edge x1 y1 && 
					  same_face x y1 && same_face y y1) in
			    let Some { rvertex = Some pt } = cube.(x1).cube_edges.(!last_dir) in
			    last_result := Some (y1, pt); true
			  with 
			    Not_found -> false)
	    in
	    let Some (y1, pt) = !last_result in
	    fn ((x1,y1)::adone) (pt::acc) x1 y1
	  with Not_found -> acc
	in 
	print_newline ();
	fn [x,y] [pt] x y
      with Not_found -> []
    in
    let rec group_edges = function
      [] -> ()
    | c::l -> 
	let rec gn = function
          [] -> ()
	| c'::l' ->
	    List.iter print_int c; print_string " ";
	    List.iter print_int c'; print_string " ";
	    let polygone = build_polygone c c' in
	    print_int (List.length polygone); print_newline (); 
	    (if List.length polygone > 2 then triangularize polygone else 
	      if polygone != [] then failwith "bug in group_edges");
	    gn l'
	in gn l; group_edges l
    in
    group_edges !composantes_lists 
  in 
  List.iter make_triangles !all_cubes;
  GlList.ends();
  list

let front_shininess =   60.0
let front_specular  =   0.7, 0.7, 0.7, 1.0
let ambient         =   0.0, 0.0, 0.0, 1.0
let diffuse         =   1.0, 1.0, 1.0, 1.0
let position0       =   1.0, 1.0, 1.0, 0.0
let position1       =   -1.0,-1.0, 1.0, 0.0
let position2       =   1.0, 1.0, -1.0, 0.0
let position3       =   -1.0,-1.0, -1.0, 0.0
let lmodel_ambient  =   0.5, 0.5, 0.5, 1.0
let lmodel_twoside  =   false

let myinit () =
  GlClear.depth 1.0;
  GlClear.color (0.0, 0.0, 0.0);
  GlDraw.color (0.8, 0.2, 0.3);
  GlDraw.shade_model `smooth;
  List.iter (GlLight.light ~num:0)
    [`ambient ambient; `diffuse diffuse; `position position0];
  List.iter (GlLight.light ~num:1)
    [`ambient ambient; `diffuse diffuse; `position position1];
  List.iter (GlLight.light ~num:2)
    [`ambient ambient; `diffuse diffuse; `position position2];
  List.iter (GlLight.light ~num:3)
    [`ambient ambient; `diffuse diffuse; `position position3];
  GlLight.light_model (`ambient lmodel_ambient);
  GlLight.light_model (`two_side lmodel_twoside);
  List.iter Gl.enable
    [`lighting;`light0;`light1;`light2;`light3;`depth_test;`normalize];

  GlLight.material ~face:`both (`shininess front_shininess);
  GlLight.material ~face:`both (`specular front_specular);

  GlMisc.hint `fog `fastest;
  GlMisc.hint `perspective_correction `fastest;
  GlMisc.hint `polygon_smooth `fastest


let my_reshape ~w ~h =
  GlDraw.viewport ~x:0 ~y:0 ~w ~h;
  GlMat.mode `projection;
  GlMat.load_identity ();
  if w <= h then
    GlMat.ortho ~x:(V.vget origin 0,V.vget origin 0+.size) ~z:(V.vget origin 2,V.vget origin 2+.size)
      ~y:(V.vget origin 1 *. float h /. float w, (V.vget origin 1+.size) *. float h /. float w)
  else 
    GlMat.ortho ~y:(V.vget origin 1,V.vget origin 1+.size) ~z:(V.vget origin 2,V.vget origin 2+.size)
      ~x:(V.vget origin 0 *. float h /. float w, (V.vget origin 0+.size) *. float h /. float w);
  GlMat.mode `modelview

(*  Main Loop
 *  Open window with initial window size, title bar, 
 *  RGBA display mode, and handle input events.
 *)

open Tk

let main () =
  let top = openTk () in
  let togl =
    Togl.create top ~double:true ~rgba:true 
      ~depth:true ~width:500 ~height:500 in
  let angle_x = ref 0.0 in
  let angle_y = ref 0.0 in
  let angle_z = ref 0.0 in
  let display surface () =
    GlClear.clear [`color; `depth];
    
    GlMat.push ();
    GlMat.rotate ~angle:!angle_x ~x:1.0 ();
    GlMat.rotate ~angle:!angle_y ~y:1.0 ();
    GlMat.rotate ~angle:!angle_z ~z:1.0 ();
    (*  solid_sphere ~radius:1.0; *)
    GlLight.material ~face:`both (`diffuse (0.7, 0.0, 0.0, 1.0));
    GlList.call surface;
    GlMat.pop ();
    Gl.flush ();
    Togl.swap_buffers togl;
  in
  Wm.title_set top "Surface";
  myinit ();
  let surface = draw_surface e in
  let step = 0.5 in
  let key_press x =
    begin
      match x with
      | "Left" | "KP_Left" -> angle_y := !angle_y -. step 
      | "Right" | "KP_right" -> angle_y := !angle_y +. step 
      | "Up" | "KP_Up" -> angle_x := !angle_x -. step 
      | "Down" | "KP_Down" -> angle_x := !angle_x +. step 
      | "KP_End" | "KP_Prior" -> angle_z := !angle_z -. step 
      | "KP_Next"| "KP_Home" -> angle_z := !angle_z +. step 
      | _ -> ()
    end;
    display surface ()
  in
  Togl.reshape_func togl
    ~cb:(fun () -> my_reshape ~w:(Togl.width togl) ~h:(Togl.height togl));
  Togl.display_func togl ~cb:(display surface);
  bind togl ~events:[`KeyPress] ~fields:[`KeySymString]
    ~action:(fun ev -> key_press ev.ev_KeySymString);
  bind togl ~events:[`Enter] ~action:(fun _ -> Focus.set togl);
  pack [togl] ~expand:true ~fill:`Both;
  mainLoop ()

let _ = Printexc.print main ()
