open Algebra
open Expression
open Vector
open Matrix
open Full_matrix
open Tk

(*
tests/geo "x^3 + y^3 + z^3 + 2*(x^2 + y^2 + z^2) - 1.1" -s 6 -a 0.3 -m 2
*)

let (++) (x,y,z) (x',y',z') = (x+.x',y+.y',z+.z')
let (+++) (x,y,z) (x',y',z') = (x+x',y+y',z+z')
let (--) (x,y,z) (x',y',z') = (x-.x',y-.y',z-.z')
let (||) (x,y,z) (x',y',z') = x*.x'+.y*.y'+.z*.z'
let norm x = sqrt (x || x)
let vprod (x,y,z) (x',y',z') = y*.z'-.y'*.z, x*.z'-.x'*.z, x*.y'-.x'*.y
let det (x,y,z) (x',y',z') (x'',y'',z'') = 
  x*.y'*.z'' +. x'*.y''*.z +. x''*.y*.z' -. x*.y''*.z' -. x'*.y*.z'' -. x''*.y'*.z
let detn a b c = 
  det a b c /. (norm a *. norm (vprod b c))
let middle (x,y,z) (x',y',z') = ((x+.x')/.2.0,(y+.y')/.2.0,(z+.z')/.2.0)
let angle x y = 
  let num = (x || y) in
  let den = sqrt ((x || x) *. (y || y)) in
  if num = 0.0 then 0.0 else
  abs_float (asin (num /. den))

module E = Expression(Field_R)

let x = E.Var "x"
let y = E.Var "y"
let z = E.Var "z"

(*
let x2 = E.Pow(x, 2)
let y2 = E.Pow(y, 2)
let z2 = E.Pow(z, 2)
let x3 = E.Pow(x, 3)
let y3 = E.Pow(y, 3)
let z3 = E.Pow(z, 3)
let sphere = E.(++) (E.( ** ) x2 (E.Cst 2.0)) 
	       (E.(++) (E.( ** ) y2 (E.Cst 2.0))  
		  (E.(--) (E.( ** ) z2 (E.Cst 2.0)) (E.Cst 1.25))) 
let cubic = E.(++) x3 (E.(++) y3 (E.(++) z3 sphere)) 
let tore = E.(--) (E.Pow(E.(++) x2 (E.(++) y2 (E.(++) z2 (E.Cst 0.75))),2))
              (E.( ** ) (E.Cst 4.0) (E.(++) x2 y2))
let planes = E.( ** ) (E.(--) (E.(++) x y) (E.Cst 1.0)) (E.(--) (E.(++) x y) (E.Cst 2.0))
let e = cubic
*)

let max_depth = 10
let stre, size, center, min_depth, verbose, max_angle  = 
  let stre = ref "x^2 + y^2 + z^2 - 1" in
  let size = ref 2.0 in
  let center = ref [| 0.0; 0.0; 0.0 |] in
  let min_depth = ref 3 in
  let verbose = ref false in
  let max_angle = ref 0.5 in
  let i = ref 0 in
  let argc = Array.length Sys.argv in
  let next () =
    incr i;
    if !i >= argc then begin
      prerr_string "Missing argument"; prerr_newline ();
      exit (1)
    end;
    Sys.argv.(!i)
  in
  while !i < argc-1 do
    incr i;
    match Sys.argv.(!i) with
      "--exp" | "-e" -> stre := next ();
    | "--verbose" | "-v" -> verbose := true
    | "--size" | "-s" -> size := float_of_string (next ())
    | "--angle" | "-a" -> max_angle := float_of_string (next ())
    | "--min" | "-m" -> min_depth := int_of_string (next ())
    | "--center" | "-c" ->
		     let cx = float_of_string (next ()) in
		     let cy = float_of_string (next ()) in
		     let cz = float_of_string (next ()) in
		     center := [| cx; cy ; cz |]
    | s -> stre := s
  done;
  !stre, !size, !center, !min_depth, !verbose, !max_angle

let str = Stream.of_string stre
let e = E.parse str
let ex = E.derive e "x" 
let ey = E.derive e "y" 
let ez = E.derive e "z" 

(* let e = sphere  *)
let hsize = size /. 2.0
let origin = (center.(0)-.hsize, center.(1)-.hsize, center.(2)-.hsize)

let eval e (x,y,z) = E.eval e ["x", x; "y", y; "z", z]  

let epsilon = 10e-10


let front_shininess =   60.0
let front_specular  =   0.7, 0.7, 0.7, 1.0
let ambient         =   0.0, 0.0, 0.0, 1.0
let diffuse         =   1.0, 1.0, 1.0, 1.0
let position0       =   1.0, 1.0, 1.0, 0.0
let position1       =   -1.0,-1.0, 1.0, 0.0
let position2       =   1.0, 1.0, -1.0, 0.0
let position3       =   -1.0,-1.0, -1.0, 0.0
let lmodel_ambient  =   0.2, 0.2, 0.2, 1.0
let lmodel_twoside  =   true

let myinit () =
  GlClear.depth 1.0;
  GlClear.color (0.0, 0.0, 0.0);
  GlDraw.shade_model `smooth;
  List.iter ~f:(GlLight.light ~num:0)
    [`ambient ambient; `diffuse diffuse; `position position0];
  List.iter ~f:(GlLight.light ~num:1)
    [`ambient ambient; `diffuse diffuse; `position position1];
  List.iter ~f:(GlLight.light ~num:2)
    [`ambient ambient; `diffuse diffuse; `position position2];
  List.iter ~f:(GlLight.light ~num:3)
    [`ambient ambient; `diffuse diffuse; `position position3];
  GlLight.light_model (`ambient lmodel_ambient);
  GlLight.light_model (`two_side lmodel_twoside); 
  List.iter ~f:Gl.enable
    [`lighting;`light0;`light1;`light2;`light3;`depth_test;`normalize];

  GlLight.material ~face:`both (`shininess front_shininess);
  GlLight.material ~face:`both (`specular front_specular);

  GlMisc.hint `fog `fastest;
  GlMisc.hint `perspective_correction `fastest;
  GlMisc.hint `polygon_smooth `fastest

let top = openTk ()
let togl =
  Togl.create top ~double:true ~rgba:true 
    ~depth:true ~width:500 ~height:500
let _ = 
  Wm.title_set top "Surface";
  myinit ()

let dighotomie f a b =
  let fa = eval f a and fb = eval f b in
  if abs_float(fa) < epsilon then a else 
  if abs_float(fb) < epsilon then b else
  let fa, a, fb, b =     
    if fa < 0.0 && fb > 0.0 then fa, a, fb, b else
    if fa > 0.0 && fb < 0.0 then fb, b, fa, a else
    raise Not_found
  in
  let rec fn fa a fb b =
    let c = middle a b in
    let fc = eval f c in 
    if abs_float(fc) < epsilon then c else
    if fc < 0.0 then fn fc c fb b else fn fa a fc c 
  in
  fn fa a fb b

type index = int * int * int
type v3 = float * float * float

let step p = 1 lsl (max_depth - p)
let max_val = step 0 

let in_box (i0,i1,i2) =
  0 <= i0 && i0 < max_val &&
  0 <= i1 && i1 < max_val &&
  0 <= i2 && i2 < max_val

type cube_vertex =
{
  index : index;
  cfval : float;
  zerov : vertex option;
  cube_edges : cube_edge option array;
  mutable gt : cube_vertex option
}

and cube_edge =
{
  cvertex1 : cube_vertex;
  cvertex2 : cube_vertex;
  rvertex : vertex option;
  depth : int;
}

and vertex = {
  pos : v3;
  normal : v3;
}

let build_vertex pos = {
  pos = pos;
  normal = (eval ex pos,eval ey pos,eval ez pos)
}

let directioni = [|
  ( 0 , 0 , 0 );
  ( 1 , 0 , 0 );
  ( 0 , 1 , 0 );
  ( 0 , 0 , 1 );
  ( 1 , 1 , 0 );
  ( 0 , 1 , 1 );
  ( 1 , 0 , 1 );
  ( 1 , 1 , 1 );
|]

let print_index v =
  let i0,i1,i2 = v.index in
  Printf.printf "(%d,%d,%d)" i0 i1 i2
  
let print_edge (x, y, e) =
  print_index x;
  print_string "-";
  print_index y;
  print_string " "

let move index dir depth =
  let i0,i1,i2 = index in
  let j0,j1,j2 = directioni.(dir) in
  (i0 + j0 * step depth, i1 + j1 * step depth, i2 + j2 * step depth)

let same_face cube i j =
  let (i0,i1,i2) = i.index in
  let (j0,j1,j2) = j.index in
  let (k0,k1,k2) = cube.(0).index in
  let (l0,l1,l2) = cube.(7).index in
  ((i0 = j0 && (i0 = k0 or i0 = l0)) or
   (i1 = j1 && (i1 = k1 or i1 = l1)) or
   (i2 = j2 && (i2 = k2 or i2 = l2)))

let same_edge i j =
  let (i0,i1,i2) = i.index in
  let (j0,j1,j2) = j.index in
  (i0 = j0 && i1 = j1) or (i0 = j0 && i2 = j2) or (i2 = j2 && i1 = j1)

let same_edges (x,y,_) (x',y',_) =
  same_edge x x' &&
  same_edge x y' &&
  same_edge y x' &&
  same_edge y y'

let direction i j =
  let (i0,i1,i2) = i.index in
  let (j0,j1,j2) = j.index in
  match i0 - j0, i1 - j1, i2 - j2 with
    x, 0, 0 when x != 0 -> if x < 0 then 0 else 1
  | 0, x, 0 when x != 0 -> if x < 0 then 2 else 3
  | 0, 0, x when x != 0 -> if x < 0 then 4 else 5
  | _ -> -1 

let common_edge i j =
  let dir = direction i j in
  if dir < 0 then raise Not_found else
    match i.cube_edges.(dir) with
      None -> raise Not_found
    | Some e -> 
	if e.cvertex1 == j or e.cvertex2 == j then e else raise Not_found

let edges = [| 0,1,0,1; 0,2,2,3; 0,3,4,5; 1,4,2,3; 1,6,4,5; 2,4,0,1;
               2,5,4,5; 3,5,2,3; 3,6,0,1; 4,7,4,5; 5,7,0,1; 6,7,2,3 |]                        
let maxpower =  1 lsl max_depth              
let position (i0,i1,i2) =
  let x0,x1,x2 = origin in
  (x0 +. size *. (float) i0 /. (float) maxpower,
   x1 +. size *. (float) i1 /. (float) maxpower,
   x2 +. size *. (float) i2 /. (float) maxpower)

let rec do_edge fn v1 v2 dir =
  let Some e = v1.cube_edges.(dir) in
  let v = 
    if e.cvertex1 == v1 then e.cvertex2 else
    if e.cvertex2 == v1 then e.cvertex1 else
    failwith "bug in do_edge"
  in
  fn e; 
  if not (v == v2) then do_edge fn v v2 dir

let rec search v = match v.gt with
    Some v' -> 
      let v'' = search v' in
      if v' != v'' then v.gt <- Some v'';
      v''
  | None -> v

let merge v w =
  let v' = search v and w' = search w in
  if v' != w' then v'.gt <- Some w'

let clean v = 
  v.gt <- None

let draw_surface e = 

  let all_vertex = Hashtbl.create 101 in

  let all_cubes = ref [] in

  let rec make_cube e index depth = 
    let fn i = 
      let index = move index i depth in
      try 
	Hashtbl.find all_vertex index
      with Not_found ->
	let pos = position index in
	let x =  eval e pos in
	let cube_vertex = { 
	  index = index; cfval = x; 
	  zerov = if x != 0.0 then None else Some (build_vertex pos); 
	  gt = None;
	  cube_edges = Array.make 6 None } 
	in
	Hashtbl.add all_vertex ~key:index ~data:cube_vertex;
	cube_vertex
    in 
    let cube = Array.init 8 ~f:fn in
    for i = 0 to 11 do
      let vi1, vi2, i1, i2 = edges.(i) in
      let v1 = cube.(vi1) and v2 = cube.(vi2) in
      match v1.cube_edges.(i1), v2.cube_edges.(i2) with
      | Some e1, Some e2 when e1.depth >= depth && e2.depth >= depth ->
	  ()
      | Some e1, Some e2 when e1.depth < depth && e2.depth >= depth ->
	  failwith "bug 1 in make_edge"
      | Some e1, Some e2 when e1.depth >= depth && e2.depth < depth ->
	  failwith "bug 2 in make_edge"
      | _ ->
	  let edge = {cvertex1 = v1; cvertex2 = v2; depth = depth; rvertex =
			let sgn x = if x >= 0.0 then 1 else -1 in
			if sgn v1.cfval * sgn v2.cfval = -1 then begin
			  if v1.zerov != None then v1.zerov else
			    if v2.zerov != None then v2.zerov else
			      let x0 = position v1.index in
			      let x1 = position v2.index in
			      let x = dighotomie e x0 x1 in
			      Some (build_vertex x)
			end
			else
			  None }
	  in v1.cube_edges.(i1) <- Some edge; v2.cube_edges.(i2) <- Some edge; 
    done;
    all_cubes := (cube,depth) :: !all_cubes 
  in
  print_string "Building initial cubes ..."; print_newline ();
  begin
    let max = (1 lsl min_depth) - 1 in
    let step = step min_depth in
    for i = 0 to max do
      for j = 0 to max do
	for k = 0 to max do
	  make_cube e (i*step,j*step,k*step) min_depth
	done
      done
    done
  end;
  let split_cube index depth =
    for i = 0 to 7 do
      let index = move index i (depth+1) in
      make_cube e index (depth + 1)
    done
  in

  print_int (List.length !all_cubes); print_string "intitial cubes."; print_newline ();

  let clean_all () = 
    Hashtbl.iter ~f:(fun ~key:_ ~data:v -> clean v) all_vertex in

  let build_classes v =
    for i = 0 to 5 do
      match v.cube_edges.(i) with 
	None -> () 
      | Some e -> 
	  let w = if e.cvertex1 == v then e.cvertex2 else e.cvertex1 in
	  if e.rvertex = None then merge v w
    done
  in

  let build_all_classes () = 
    Hashtbl.iter ~f:(fun ~key:_ ~data:v -> build_classes v) all_vertex 
  in

  let vertex a =
    GlDraw.normal3 (a.normal);
    GlDraw.vertex3 (a.pos)
  in
  let count = ref 0 in
  let continue = ref false in
  let mark = ref false in
  let triangle depth a b c =
    if a <> b && a <> c && b <> c then begin 
      incr count;
      let da = detn a.normal (b.pos -- a.pos) (c.pos -- a.pos) in
      let db = detn b.normal (c.pos -- b.pos) (a.pos -- b.pos) in
      let dc = detn c.normal (a.pos -- c.pos) (b.pos -- c.pos) in
      if not (da *. db >= 0.0 &&  da *. dc >= 0.0 && 
	(1.0 -. abs_float da) /.(float depth) <= max_angle &&
	(1.0 -. abs_float db) /. (float depth) <= max_angle &&
	(1.0 -. abs_float dc) /. (float depth) <= max_angle) 
      then mark := true;
      GlDraw.begins `triangle_strip;
      if da >= 0.0 then (vertex a; vertex b; vertex c)
      else  (vertex a; vertex c; vertex b);
      GlDraw.ends ()
    end
  in
  let rec triangularize depth l =
    match l with
      x::y::z::l -> triangle depth x y z; triangularize depth (x::z::l); triangularize depth (z::l@[x])
    | _ -> ()
  in

  let special_polygones = Hashtbl.create 101 in

  let make_triangles depth cube =
    let (i0, i1, i2) = cube.(0).index in
    let (j0, j1, j2) = cube.(7).index in
    let in_cube c =
      let (k0, k1, k2) = c.index in
      i0 <= k0 && k0 <= j0 && i1 <= k1 && k1 <= j1 && i2 <= k2 && k2 <= j2
    in
    let composantes = ref [] in
    let adone = ref [] in
    let rec collect c =
      if in_cube c && not (List.memq c !adone) then begin
	adone := c::!adone;
	if c.cfval < 0.0 then 
	  composantes := c::!composantes;
	for i = 0 to 5 do
	  match c.cube_edges.(i) with 
	    None -> () 
	  | Some e -> 
	      if c == e.cvertex1 then collect e.cvertex2 
	      else collect e.cvertex1
	done
      end
    in
    collect cube.(0);
    if verbose then begin
      print_string "Cubes: ";
      List.iter ~f:print_index !composantes;
      print_newline ()
    end;
    let composantes_lists = ref [] in
    let group_classes v =
      let cur = ref [] in
      let v' = search v in
      if not (List.memq v !cur) then begin
	List.iter ~f:(fun w -> if search w == v' then cur := w::!cur)
	  !composantes;
	composantes_lists := !cur::!composantes_lists
      end
    in
    List.iter ~f:group_classes !composantes;
    let search l cond =
      let rec fn acc = function
          y::l -> 
	    if cond y then y, acc@l
	    else fn (y::acc) l
	| [] -> raise Not_found
      in fn [] l
    in
    let find_edges c =
      let acc = ref [] in
      let rec fn = function
	  [] -> !acc
	| x::l ->
	    for i = 0 to 5 do
	      match x.cube_edges.(i) with
		None -> ()
	      | Some e ->
		  let y = 
		    if x == e.cvertex1 then e.cvertex2 else e.cvertex1 in
		  if e.rvertex <> None && in_cube y then acc:=(x,y,e)::!acc
	    done;
	    fn l
      in fn c
    in
    let build_polygone c = 
      if verbose then begin
	print_string "Composantes: ";
	List.iter ~f:print_index c;
	print_newline ()
      end;
      let edges = find_edges c in
      if verbose then begin
	print_string "Arrtes: ";
	List.iter ~f:print_edge edges;
	print_newline ();
	flush stdout
      end;
      let max4 a b c d =
	max (max a b) (max c d)
      in
      let min4 a b c d =
	min (min a b) (min c d)
	  in
      let no_included x y x' y' =
	let x0,x1,x2 = x.index in
	let y0,y1,y2 = y.index in
	let x0',x1',x2' = x'.index in
	let y0',y1',y2' = y'.index in
	let min0 = min4 x0 y0 x0' y0' in
	let min1 = min4 x1 y1 x1' y1' in
	let min2 = min4 x2 y2 x2' y2' in
	let max0 = max4 x0 y0 x0' y0' in
	let max1 = max4 x1 y1 x1' y1' in
	let max2 = max4 x2 y2 x2' y2' in
	let included (x'',y'',_) = 
	  let x0,x1,x2 = x''.index in
	  let y0,y1,y2 = y''.index in
	  let f min x max =
	    if min > x or max < x then raise Not_found else
	      if min < x && x < max then 1 else 0
	  in
	  try
	    if 
	      (x != x'' or y != y'') && (x' != x'' or y' != y'') &&
	      (f min0 x0 max0 + f min1 x1 max1 + f min2 x2 max2 +
	       f min0 y0 max0 + f min1 y1 max1 + f min2 y2 max2) > 0
	      (* non equivalent  or  cause de l'exception ! *)
	    then raise Exit
	  with Not_found -> () 
	in
	try
	  List.iter ~f:included edges; true
	with Exit -> false
      in
      let compat (x,y,e) (x',y',e') =
	same_face cube x x' && same_face cube y y' && 
	same_face cube x y' && same_face cube x' y &&
	no_included x y x' y'
      in
      let rec fn acca acc l =
	match l with 
	  [] -> acc::acca 
	| e::l ->
	    try 
	      let e',l' = search l (compat e) in
	      fn acca (e::acc) (e'::l')
	    with Not_found ->
	      fn ((e::acc)::acca) [] l
      in
      fn [] [] edges
    in
    let rec add_to_special (e1: cube_vertex * cube_vertex * cube_edge) e2 p =
      print_string "#";
      let key, data = if e1 <= e2 then (e1, e2), p else (e2, e1), List.rev p in
      try 
	let p' = Hashtbl.find special_polygones key in
	Hashtbl.remove special_polygones key;
	let p'' = e1::List.rev p@e2::p' in
	is_special p''
      with Not_found -> 
	Hashtbl.add special_polygones ~key ~data;
	[]
    and is_special = function []| [_]|[_;_] -> [] | e0::l as ori ->
      let rec fn acc e = function
	| [e'] -> 
	    if same_edges e0 e' then add_to_special e' e0 (List.rev acc)
	    else ori
	| e'::l -> 
	    if same_edges e e' then add_to_special e e' (l@List.rev acc@[e0]) 
	    else fn (e::acc) e' l
      in fn [] e0 l
    in 
    let extract p = 
      match p with 
	[] -> [] 
      | [_] | [_;_] -> print_string "*"; []
      | l ->
	  List.map ~f:(function _,_,{rvertex = Some x} -> x) l
    in	
    let rec group_edges acc = function
      [] -> acc
    | c::l -> 
	let polygones = build_polygone c in
	if verbose then begin
	  let fn p = List.iter ~f:print_edge p;
	    print_newline ();
	  in
	  print_string "Polygnes: ";
	  List.iter ~f:fn polygones;
	  print_newline ()
	end;
	let polygones = List.map ~f:is_special polygones in
	let vertexes = List.map ~f:extract polygones in
	group_edges (vertexes@acc) l
    in 
    let polygones = group_edges [] !composantes_lists in
    List.iter ~f:(triangularize depth) polygones;
  in 
  let gllist = ref (GlList.create `compile_and_execute) in
  let generation = ref 0 in
  let continue = ref true in
  let to_split = ref [] in
  let one_loop ((cube, depth) as c : (cube_vertex array * int)) =
    mark := false;
    make_triangles depth cube;
    if !mark then begin
      continue := true;
      to_split := (cube.(0).index, depth)::!to_split
    end else
      all_cubes:=c::!all_cubes
  in
  while !continue do
    GlClear.clear [`color; `depth];
    continue := false;
    count := 0;
    incr generation;
    build_all_classes ();
    print_string "Building triangles generation "; 
    print_int !generation; print_newline ();
    let l = !all_cubes in
    all_cubes := [];
    List.iter ~f:one_loop l;
    List.iter ~f:(fun (i, d) -> split_cube i d) !to_split;
    to_split := [];
    clean_all ();
    GlList.ends ();
    let c = ref 0 in
    Hashtbl.iter ~f:(fun ~key ~data -> incr c)  special_polygones;
    Hashtbl.clear special_polygones;
    print_int !c; print_string " remaining special polygones."; print_newline ();
    print_int !count; print_string " triangles."; print_newline ();
    print_int (List.length !all_cubes); print_string " cubes."; print_newline ();
    Togl.swap_buffers togl;
    if !continue then begin
      GlList.delete !gllist;
      gllist := GlList.create `compile_and_execute
    end;
  done;    
  print_string "Done";
  print_newline ();

  !gllist


let my_reshape ~w ~h =
  GlDraw.viewport ~x:0 ~y:0 ~w ~h;
  GlMat.mode `projection;
  GlMat.load_identity ();
  let x0,x1,x2 = origin in
  if w <= h then
    GlMat.ortho ~x:(x0,x0+.size) ~z:(x2-.size,x2+.2.0*.size)
      ~y:(x1 *. float h /. float w, (x1+.size) *. float h /. float w)
  else 
    GlMat.ortho ~y:(x1,x1+.size) ~z:(x2-.size,x2+.2.0*.size)
      ~x:(x0 *. float h /. float w, (x0+.size) *. float h /. float w);
  GlMat.mode `modelview

(*  Main Loop
 *  Open window with initial window size, title bar, 
 *  RGBA display mode, and handle input events.
 *)

let main () =
  let angle_x = ref 0.0 in
  let angle_y = ref 0.0 in
  let angle_z = ref 0.0 in
  let surface = ref None in
  let display surface () =
    GlClear.clear [`color; `depth];
    GlDraw.color  (1.0, 1.0, 1.0);
   
    GlMat.push ();
    GlMat.rotate ~angle:!angle_x ~x:1.0 ();
    GlMat.rotate ~angle:!angle_y ~y:1.0 ();
    GlMat.rotate ~angle:!angle_z ~z:1.0 ();

    GlLight.material ~face:`front (`diffuse (0.7, 0.0, 0.0, 1.0));
    GlLight.material ~face:`back (`diffuse (0.0, 0.7, 0.0, 1.0));
    (match !surface with
      None -> surface := Some (draw_surface e)
    | Some s -> GlList.call s; Togl.swap_buffers togl);
    GlMat.pop ();
    Gl.flush ();
  in
  let step = 1.0 in
  let key_press x =
    begin
      match x with
      | "Left" | "KP_Left" -> angle_y := !angle_y -. step 
      | "Right" | "KP_right" -> angle_y := !angle_y +. step 
      | "Up" | "KP_Up" -> angle_x := !angle_x -. step 
      | "Down" | "KP_Down" -> angle_x := !angle_x +. step 
      | "KP_End" | "KP_Prior" -> angle_z := !angle_z -. step 
      | "KP_Next"| "KP_Home" -> angle_z := !angle_z +. step 
      | _ -> ()
    end;
    display surface ()
  in
  Togl.reshape_func togl
    ~cb:(fun () -> my_reshape ~w:(Togl.width togl) ~h:(Togl.height togl));
  Togl.display_func togl ~cb:(display surface);
  bind togl ~events:[`KeyPress] ~fields:[`KeySymString]
    ~action:(fun ev -> key_press ev.ev_KeySymString);
  bind togl ~events:[`Enter] ~action:(fun _ -> Focus.set togl);
  pack [togl] ~expand:true ~fill:`Both;
  mainLoop ()

let _ = Printexc.print main ()
