open Format
open Algebra
open Expression
open Tk

module E = Expression(Field_R)

type point = float * float * float

type vector = point

let (++) (x,y,z) (x',y',z') = (x+.x', y+.y', z+.z')

let (--) (x,y,z) (x',y',z') = (x-.x',y-.y',z-.z')

let (||) (x,y,z) (x',y',z') = x*.x'+.y*.y'+.z*.z'

let (//) (x,y,z) t = (x/.t, y/.t, z/.t)

let ( ** ) (x,y,z) t = (x*.t, y*.t, z*.t)

let printv (x,y,z) = 
  print_string "(";
  print_float x; print_space ();
  print_float y; print_space ();
  print_float z; print_string ")"

let norm x = sqrt (x || x)

let vprod (x,y,z) (x',y',z') = y*.z'-.y'*.z, x'*.z-.x*.z', x*.y'-.x'*.y

let det (x,y,z) (x',y',z') (x'',y'',z'') = 
  x*.y'*.z'' +. x'*.y''*.z +. x''*.y*.z' -. 
  x*.y''*.z' -. x'*.y*.z'' -. x''*.y'*.z

let front_shininess =   60.0
let front_specular  =   0.7, 0.7, 0.7, 1.0
let ambient         =   0.0, 0.0, 0.0, 1.0
let diffuse         =   1.0, 1.0, 1.0, 1.0
let position0       =   1.0, 1.0, 1.0, 0.0
let position1       =   -1.0,-1.0, 1.0, 0.0
let position2       =   1.0, 1.0, -1.0, 0.0
let position3       =   -1.0,-1.0, -1.0, 0.0
let lmodel_ambient  =   0.2, 0.2, 0.2, 1.0
let lmodel_twoside  =   true

let myinit () =
  GlClear.depth 1.0;
  GlClear.color (0.0, 0.0, 0.0);
  GlDraw.shade_model `smooth;
  List.iter (GlLight.light ~num:0)
    [`ambient ambient; `diffuse diffuse; `position position0];
  List.iter (GlLight.light ~num:1)
    [`ambient ambient; `diffuse diffuse; `position position1];
  List.iter (GlLight.light ~num:2)
    [`ambient ambient; `diffuse diffuse; `position position2];
  List.iter (GlLight.light ~num:3)
    [`ambient ambient; `diffuse diffuse; `position position3];
  GlLight.light_model (`ambient lmodel_ambient);
  GlLight.light_model (`two_side lmodel_twoside); 
  List.iter Gl.enable
    [`lighting;`light0;`light1;`light2;`light3;`depth_test;`normalize];

  GlLight.material ~face:`both (`shininess front_shininess);
  GlLight.material ~face:`both (`specular front_specular);

  GlMisc.hint `fog `fastest;
  GlMisc.hint `perspective_correction `fastest;
  GlMisc.hint `polygon_smooth `fastest

let top = openTk ()
let togl =
  Togl.create top ~double:true ~rgba:true 
    ~depth:true ~width:500 ~height:500
let _ = 
  Wm.title_set top "Surface";
  myinit ()

type vertex = {
    mutable position : point;
    mutable value : float;
    mutable grad : vector;
    mutable triangles : triangle list;
    mutable on_surface : bool;
    mutable blocked : bool;
    mutable move : vector;
  } 

and triangle =
  {
    (* orientation matters ! *)
    vertex1 : vertex; 
    vertex2 : vertex; 
    vertex3 : vertex;
    mutable neighbour1 : triangle option 
      (* triangle between vertex2 and vertex3 *);
    mutable neighbour2 : triangle option 
      (* triangle between vertex1 and vertex3 *); 
    mutable neighbour3 : triangle option 
      (* triangle between vertex1 and vertex2 *);
    mutable divided : (vertex * vertex * vertex) option;
    mutable flag1 : bool;
 } 

let normal_triangle t =
  let v = vprod (t.vertex2.position -- t.vertex1.position) 
                (t.vertex3.position -- t.vertex1.position) in
  v // norm v

let surface_triangle t =
  let x1 = t.vertex2.position -- t.vertex1.position in
  let x2 = t.vertex3.position -- t.vertex1.position in
  let p = norm x1 *. norm x2 in
  let s = x1 || x2 in
  p -. s*.s /. p

let glDraw_vertex v =
  GlDraw.normal3 (v.grad);
  GlDraw.vertex3 (v.position)

let glDraw_triangle t =
  GlDraw.begins `triangle_strip;
  glDraw_vertex t.vertex1;
  glDraw_vertex t.vertex2;
  glDraw_vertex t.vertex3;
  GlDraw.ends ()
  
let glDraw_triangles ts =
   let gllist = GlList.create `compile_and_execute in
   List.iter glDraw_triangle ts;
   GlList.ends ();
   gllist

let origin = -1.5,-1.5,-1.5
let size = 3.0

let my_reshape ~w ~h =
  GlDraw.viewport ~x:0 ~y:0 ~w ~h;
  GlMat.mode `projection;
  GlMat.load_identity ();
  let x0,x1,x2 = origin in
  if w <= h then
    GlMat.ortho ~x:(x0,x0+.size) ~z:(x2-.size,x2+.2.0*.size)
      ~y:(x1 *. float h /. float w, (x1+.size) *. float h /. float w)
  else 
    GlMat.ortho ~y:(x1,x1+.size) ~z:(x2-.size,x2+.2.0*.size)
      ~x:(x0 *. float h /. float w, (x0+.size) *. float h /. float w);
  GlMat.mode `modelview

let glLoop update =
  let gllist = ref (update ()) in
  let angle_x = ref 0.0 in
  let angle_y = ref 0.0 in
  let angle_z = ref 0.0 in

  let display () =
    GlClear.clear [`color; `depth];
    GlDraw.color  (1.0, 1.0, 1.0);
   
    GlMat.push ();
    GlMat.rotate ~angle:!angle_x ~x:1.0 ();
    GlMat.rotate ~angle:!angle_y ~y:1.0 ();
    GlMat.rotate ~angle:!angle_z ~z:1.0 ();

    GlLight.material ~face:`front (`diffuse (0.7, 0.0, 0.0, 1.0));
    GlLight.material ~face:`back (`diffuse (0.0, 0.7, 0.0, 1.0));
    GlList.call !gllist; Togl.swap_buffers togl;
    GlMat.pop ();
    Gl.flush ();
  in
  let step = 1.0 in
  let key_press x =
    begin
      match x with
      | "Left" | "KP_Left" -> angle_y := !angle_y -. step 
      | "Right" | "KP_right" -> angle_y := !angle_y +. step 
      | "Up" | "KP_Up" -> angle_x := !angle_x -. step 
      | "Down" | "KP_Down" -> angle_x := !angle_x +. step 
      | "KP_End" | "KP_Prior" -> angle_z := !angle_z -. step 
      | "KP_Next"| "KP_Home" -> angle_z := !angle_z +. step
      |	 "U" | "u" ->
          GlList.delete !gllist;
	  gllist := update ();
      | _ -> ()
    end;
    display ()
  in
  Togl.reshape_func togl
    ~cb:(fun () -> my_reshape ~w:(Togl.width togl) ~h:(Togl.height togl));
  Togl.display_func togl ~cb:(display);
  bind togl ~events:[`KeyPress] ~fields:[`KeySymString]
    ~action:(fun ev -> key_press ev.ev_KeySymString);
  bind togl ~events:[`Enter] ~action:(fun _ -> Focus.set togl);
  pack [togl] ~expand:true ~fill:`Both;
  mainLoop ()
 
module type Surface =
  sig
    val f : E.elem
  end

module Surface_polygonizer = functor (F : Surface) ->
  struct

    let f = F.f

    let fx = E.derive f "x" and fy = E.derive f "y" and fz = E.derive f "z"

    let eval' f (x,y,z) = E.eval f ["x",x; "y",y; "z",z]

    let eval p = eval' f p

    let grad p = 
      (eval' fx p, eval' fy p, eval' fz p)

    let null = 0.0,0.0,0.0

    let create_vertex p = 
      let v = eval p in
      { 
	position = p;
	value = v;
	grad = grad p;
	triangles = [];
	on_surface = v = 0.0;
	blocked = false;
	move = null;
      }	

    let create_vertex_on_surface p =
      let v = create_vertex p in
      v.on_surface <- true;
      v

    let rec inter l1 l2 = match l1, l2 with
      [], l2 -> None
    | (a::l1), l2 -> if List.mem a l2 then Some a else inter l1 l2
	
    let create_triangle v1 v2 v3 =
      if v1 = v2 or v1 = v3 or v2 = v3 then raise Exit;
      let t = 
	{
	 vertex1 = v1; vertex2 = v2; vertex3 = v3;
	 neighbour1 = inter v2.triangles v3.triangles;
	 neighbour2 = inter v1.triangles v3.triangles;
	 neighbour3 = inter v1.triangles v2.triangles;
	 divided = None;
	 flag1 = false;
       } in
      v1.triangles <- t::v1.triangles;
      v2.triangles <- t::v2.triangles;
      v3.triangles <- t::v3.triangles;
      let fix ngh vi vj =
	let t = Some t in
	match ngh with
	  None -> ()
	| Some ngh ->
	    if ((ngh.vertex1 = vi) && (ngh.vertex2 = vj)) or
	    ((ngh.vertex1 = vj) && (ngh.vertex2 = vi)) then
	      ngh.neighbour3 <- t;
	    if ((ngh.vertex1 = vi) && (ngh.vertex3 = vj)) or
	    ((ngh.vertex1 = vj) && (ngh.vertex3 = vi)) then 
	      ngh.neighbour2 <- t;
	    if ((ngh.vertex2 = vi) && (ngh.vertex3 = vj)) or
	    ((ngh.vertex2 = vj) && (ngh.vertex3 = vi)) then 
	      ngh.neighbour1 <- t
      in
      fix t.neighbour1 v2 v3;
      fix t.neighbour2 v1 v3;
      fix t.neighbour3 v1 v2;
      t 


    let make_cube origin size =
      let v000 = create_vertex (origin ++ (0.0,0.0,0.0)) in
      let v001 = create_vertex (origin ++ (0.0,0.0,size)) in
      let v010 = create_vertex (origin ++ (0.0,size,0.0)) in
      let v100 = create_vertex (origin ++ (size,0.0,0.0)) in
      let v110 = create_vertex (origin ++ (size,size,0.0)) in
      let v011 = create_vertex (origin ++ (0.0,size,size)) in
      let v101 = create_vertex (origin ++ (size,0.0,size)) in
      let v111 = create_vertex (origin ++ (size,size,size)) in
      
      [
       create_triangle v000 v001 v010;
       create_triangle v010 v001 v011;
       
       create_triangle v001 v101 v011;
       create_triangle v011 v101 v111;
       
       create_triangle v101 v110 v111;
       create_triangle v100 v110 v101;
       
       create_triangle v110 v100 v010;
       create_triangle v010 v100 v000;
       
       create_triangle v010 v011 v110;
       create_triangle v110 v011 v111;
       
       create_triangle v000 v100 v001;
       create_triangle v001 v100 v101;
     ] 

    (* split a triangle in four by the middle of the edges,
       may introduce crack if you do not split all triangles *)
    let divide_edges middle t =
      t.flag1 <- false;
      let divide_edge ngh vi vj =
	match ngh with 
	  None -> middle vi vj
	| Some ngh ->
	    match ngh.divided with
	      None -> middle vi vj
	    | Some (vd1,vd2,vd3) ->
		if ((ngh.vertex1 = vi) && (ngh.vertex2 = vj)) or
		  ((ngh.vertex1 = vj) && (ngh.vertex2 = vi)) then
		  vd3 
		else if ((ngh.vertex1 = vi) && (ngh.vertex3 = vj)) or
		  ((ngh.vertex1 = vj) && (ngh.vertex3 = vi)) then 
		  vd2 
		else if ((ngh.vertex2 = vi) && (ngh.vertex3 = vj)) or
		  ((ngh.vertex2 = vj) && (ngh.vertex3 = vi)) then 
		  vd1
		else failwith "bug in divide_edges"
      in
      let v1' = divide_edge t.neighbour1 t.vertex2 t.vertex3 in
      let v2' = divide_edge t.neighbour2 t.vertex1 t.vertex3 in
      let v3' = divide_edge t.neighbour3 t.vertex1 t.vertex2 in
      let fd t' = t <> t' in
      t.vertex1.triangles <- List.filter fd t.vertex1.triangles;
      t.vertex2.triangles <- List.filter fd t.vertex2.triangles;
      t.vertex3.triangles <- List.filter fd t.vertex3.triangles;

      t.divided <- Some (v1',v2',v3');
      [
       create_triangle v1' v2' v3';
       create_triangle t.vertex1 v3' v2';
       create_triangle t.vertex2 v1' v3';
       create_triangle t.vertex3 v2' v1';
     ] 

    (* fix the crack of a triangle *)
    let adjust_divide t =
      if not t.flag1 then [t] else begin
	t.flag1 <- false;
	let search_edge ngh vi vj =
	  match ngh with 
	    None -> None
	  | Some ngh ->
	      match ngh.divided with
		None -> None
	      | Some (vd1,vd2,vd3) ->
		  if ((ngh.vertex1 = vi) && (ngh.vertex2 = vj)) or
		    ((ngh.vertex1 = vj) && (ngh.vertex2 = vi)) then
		    Some vd3 
		  else if ((ngh.vertex1 = vi) && (ngh.vertex3 = vj)) or
		    ((ngh.vertex1 = vj) && (ngh.vertex3 = vi)) then 
		    Some vd2 
		  else if ((ngh.vertex2 = vi) && (ngh.vertex3 = vj)) or
		    ((ngh.vertex2 = vj) && (ngh.vertex3 = vi)) then 
		    Some vd1
		  else failwith "bug in adjust_divide"
	in
	let v1' = search_edge t.neighbour1 t.vertex2 t.vertex3 in
	let v2' = search_edge t.neighbour2 t.vertex1 t.vertex3 in
	let v3' = search_edge t.neighbour3 t.vertex1 t.vertex2 in
	if v1' = None && v2' = None && v3' = None then
	  failwith "bug2 in adjust_divide";
	let fd t' = t <> t' in
	t.vertex1.triangles <- List.filter fd t.vertex1.triangles;
	t.vertex2.triangles <- List.filter fd t.vertex2.triangles;
	t.vertex3.triangles <- List.filter fd t.vertex3.triangles;
	let v1' = match v1' with None -> t.vertex2 | Some v -> v in
	let v2' = match v2' with None -> t.vertex3 | Some v -> v in
	let v3' = match v3' with None -> t.vertex1 | Some v -> v in
        let r = try create_triangle v1' v2' v3'::[] with Exit -> [] in
        let r = try create_triangle t.vertex1 v3' v2'::r with Exit -> r in
        let r = try create_triangle t.vertex2 v1' v3'::r with Exit -> r in
        let r = try create_triangle t.vertex3 v2' v1'::r with Exit -> r in
	r
      end

    let divide_edges_middle = divide_edges
	(fun vi vj -> create_vertex ((vi.position ++ vj.position) // 2.0))

    let epsilon = 10e-8 
    let dighotomie a fa b fb =
      if abs_float(fa) < epsilon then a else 
      if abs_float(fb) < epsilon then b else
      let fa, a, fb, b =     
	if fa < 0.0 && fb > 0.0 then fa, a, fb, b else
	if fa > 0.0 && fb < 0.0 then fb, b, fa, a else
	raise Not_found
      in
      let rec fn fa a fb b =
	let c = (a ++ b) // 2.0 in
	let fc = eval c in 
	if abs_float(fc) < epsilon then c else
	if fc < 0.0 then fn fc c fb b else fn fa a fc c 
      in
      fn fa a fb b

    let divide_edges_digho = divide_edges
	(fun vi vj -> 
	  if vi.value *. vj.value > 0.0 then
	    create_vertex ((vi.position ++ vj.position) // 2.0)
	  else
	    create_vertex_on_surface
	      (dighotomie vi.position vi.value
		 vj.position vj.value))


    let init_depth = 4
    let ts = make_cube (-1.0,-1.0,-1.0) 2.0 
    let rec init_divide n ts =
      if n = 0 then ts else
      init_divide (n-1) (List.flatten (List.map divide_edges_middle ts))
    let ts = init_divide init_depth ts 

    let select_triangles ts =
      let bad vertex =
        vertex.on_surface or vertex.value < 0.0
      in
      let good vertex =
        vertex.on_surface or vertex.value > 0.0
      in
      let rec test simple t = 
	if bad t.vertex1 && bad t.vertex2 && bad t.vertex3 then
	  []
	else if good t.vertex1 && good t.vertex2 && good t.vertex3 then
	  [t]
	else begin
	  if simple then failwith "bug in select triangles";
	  let setflag1 = function
	      None -> ()
	    | Some t -> t.flag1 <- true
	  in
	  setflag1 t.neighbour1;
	  setflag1 t.neighbour2;
	  setflag1 t.neighbour3;
	  List.flatten (List.map (test true) (divide_edges_digho t))
	end
      in
      List.flatten (List.map adjust_divide  
		      (List.flatten (List.map (test false) ts)))
      
    

    let ts = select_triangles ts 

    let _ = List.iter (function t ->
	if t.vertex1.triangles <> [] then t.vertex1.triangles <- [];
	if t.vertex2.triangles <> [] then t.vertex2.triangles <- [];
	if t.vertex3.triangles <> [] then t.vertex3.triangles <- []) ts
 
    let rec collect vertexes = function
	[] -> vertexes
      |	t::ts ->
	  let vertexes = if t.vertex1.triangles = [] then 
	    t.vertex1::vertexes else vertexes 
	  in
	  let vertexes = if t.vertex2.triangles = [] then 
	    t.vertex2::vertexes else vertexes 
	  in
	  let vertexes = if t.vertex3.triangles = [] then 
	    t.vertex3::vertexes else vertexes 
	  in
	  t.vertex1.triangles <- t::t.vertex1.triangles;
	  t.vertex2.triangles <- t::t.vertex2.triangles;
	  t.vertex3.triangles <- t::t.vertex3.triangles;
	  collect vertexes ts
    

    let vs = collect [] ts

    let _ = 
      print_string "number of vertexes : ";
      print_int (List.length vs); print_newline ();
      print_string "number of triangles : ";
      print_int (List.length ts); print_newline ()

    let alpha = 0.5
    let beta = -1.0

    let do_one_loop vs =
      let fn v =
(*
	let fn x t = x ++ normal_triangle t ** surface_triangle t in
	let fout = List.fold_left fn null v.triangles in
	let s = norm fout in
        let nfout = fout // s in
        let fout = (nfout ** sqrt(s)) ** alpha in
	let s = norm fout in
	let fin = beta *. v.value /. (v.grad || nfout)  in
*)
	v.move <- v.grad ** (beta *. v.value /. norm v.grad /. norm v.grad);
(*
	printv v.position; print_space ();
	printv v.move; print_space ();
	printv nfout; print_space ();
	print_float v.value; print_space ();
	printv v.grad; print_newline ()
*)
      in
      List.iter fn vs;
      let gn v =
	if not v.on_surface then begin
	  v.position <- v.position ++ v.move;
          v.value <- eval v.position;
	  v.grad <- grad v.position;
	  if v.value < epsilon then v.on_surface <- true
	end
      in
      List.iter gn vs

    let update () =
      let r = glDraw_triangles ts in
      do_one_loop vs; 
      r

    let _ = 
      glLoop update


end

module F = struct
  (* tore *)
(*  let fstring = "- 4*(x^2+y^2) + ((x^2+y^2)^2 + 1 + z^2 + 1/16)^2" *)
(*  let fstring = "x^3+y^3+z^3"  *)
  let fstring = "8*x^3 + 8*y^3 + 8*z^3 + 8*(x^2 + y^2 + z^2) - 1.1"
  let str = Stream.of_string fstring
  let f = E.parse str
end

module Work = Surface_polygonizer (F)

