open Format
open Algebra
open Expression
open Vector
open Matrix
open Full_matrix

(* on travaille en dimension 3 *)
let dim = 3 
module Dim = struct let dim = dim end

(* Les vecteur de R^3 *)
module V = Vector.Make_Normed(Field_R)(Dim)

(* Les matrices 3x3 *)
module M = Make_Square(Field_R)(V)

(* Les expressions  valeur de R *)
module E = Expression(Field_R)

(* Les champs de vecteurs *)
module VF = Vector.Make(E)(Dim)

(* Les champs de matrices *)
module MF = Make_Square(E)(VF)

let grad f =
  let dfx = E.derive f "x" in
  let dfy = E.derive f "y" in
  let dfz = E.derive f "z" in
  VF.from_array [| dfx; dfy; dfz |]

let lapl f =
  let grad_f = grad f in
  E.Add(E.Add(E.Pow(VF.vget grad_f 0, 2),E.Pow(VF.vget grad_f 1, 2)),E.Pow(VF.vget grad_f 2, 2))

let rec sqrt = {E.name = "sqrt"; E.calcul = Pervasives.sqrt; 
	        E.deriv = E.Div (E.Cst(-0.5), E.Tra(sqrt, (E.Var "x")));
                E.name_var = "x"}

let hess f =
  let grad_f = grad f in
  let init setup =
    let exx = E.derive (VF.vget grad_f 0) "x" in
    setup 0 0 exx;
    let eyy = E.derive (VF.vget grad_f 1) "y" in
    setup 1 1 eyy;
    let ezz = E.derive (VF.vget grad_f 2) "z" in
    setup 2 2 ezz;
    let exy = E.derive (VF.vget grad_f 0) "y" in
    setup 0 1 exy;
    setup 1 0 exy;
    let exz = E.derive (VF.vget grad_f 0) "z" in
    setup 0 2 exz;
    setup 2 0 exz;
    let eyz = E.derive (VF.vget grad_f 1) "z" in
    setup 1 2 eyz;
    setup 2 1 eyz;
  in MF.create init
	
let eval f x =
  let xs = ["x", V.vget x 0; "y", V.vget x 1; "z", V.vget x 2] in
  E.eval f xs

let eval_grad v x = 
  let xs = ["x", V.vget x 0; "y", V.vget x 1; "z", V.vget x 2] in
  V.from_array [|
      E.eval (VF.vget v 0) xs;
      E.eval (VF.vget v 1) xs;
      E.eval (VF.vget v 2) xs
  |]

module Det = struct
  open E 

  let det m = 
  (MF.get m 0 0)**(MF.get m 1 1)**(MF.get m 2 2) ++ 
  (MF.get m 0 1)**(MF.get m 1 2)**(MF.get m 2 0) ++ 
  (MF.get m 0 2)**(MF.get m 1 0)**(MF.get m 2 1) -- 
  (MF.get m 0 0)**(MF.get m 1 2)**(MF.get m 2 1) -- 
  (MF.get m 0 1)**(MF.get m 1 0)**(MF.get m 2 2) -- 
  (MF.get m 0 2)**(MF.get m 1 1)**(MF.get m 2 0)
end

let det = Det.det
 
let curv f =
  E.simplify (E.(//) (det (hess f)) (E.( ** ) (E.Cst 2.0) (lapl f)))
  
let epsilon = 10e-10

let middle a b = V.map2 (fun x y -> (x +. y) /. 2.0) a b 

let dighotomie f a b =
  let fa = eval f a and fb = eval f b in
  if abs_float(fa) < epsilon then a else 
  if abs_float(fb) < epsilon then b else
  let fa, a, fb, b =     
    if fa < 0.0 && fb > 0.0 then fa, a, fb, b else
    if fa > 0.0 && fb < 0.0 then fb, b, fa, a else
    raise Not_found
  in
  let rec fn fa a fb b =
    let c = middle a b in
    let fc = eval f c in 
    if abs_float(fc) < epsilon then c else
    if fc < 0.0 then fn fc c fb b else fn fa a fc c 
  in
  fn fa a fb b

